﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace RfidConsoleApp
{
    public class UdpHelper
    {
        /// <summary>
        /// 接收数据 注册事件
        /// </summary>
        public Action<byte[], EndPoint> ReceiveDataAction;
        /// <summary>
        ///  是否开启接收udp数据线程flag
        /// </summary>
        public bool IsUdpcRecvStart = false;
        /// <summary>
        /// udp
        /// </summary>
        private UdpClient udpclient = null;
        /// <summary>
        /// 初始化
        /// </summary>
        /// <param name="port"></param>
        public bool Init(int port)
        {
            try
            {
                udpclient = new UdpClient(port);
                //udpclient.EnableBroadcast = true;
                return true;
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
                udpclient = null;
                return false;
            }

        }
        /// <summary>
        /// 初始化
        /// </summary>
        /// <param name="ip"></param>
        /// <param name="port"></param>
        public bool Init(string ip, int port)
        {
            try
            {
                IPEndPoint localpt = new IPEndPoint(IPAddress.Parse(ip), port);
                udpclient = new UdpClient(localpt);
                return true;
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
                udpclient = null;
                return false;
            }
        }
        /// <summary>
        /// 发送消息
        /// </summary>
        /// <param name="bytes"></param>
        /// <param name="ip"></param>
        /// <param name="port"></param>
        public void Send(Byte[] bytes, IPEndPoint RemoteEndPoint)
        {
            try
            {
                if (udpclient == null)
                {
                    return;
                }
                udpclient.BeginSend(bytes, bytes.Length, RemoteEndPoint, SendCallback, null);
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
                return;
            }
        }
        /// <summary>
        /// 发送消息回调
        /// </summary>
        /// <param name="ar"></param>
        private void SendCallback(IAsyncResult ar)
        {
            int sendCount = udpclient.EndSend(ar);
            if (sendCount == 0)
            {
                Console.WriteLine("Send a message failure...");
            }
        }
        /// <summary>
        /// 关闭接收数据线程
        /// </summary>
        public void CloseReceive()
        {
            if (IsUdpcRecvStart)
            {
                IsUdpcRecvStart = false;
                udpclient.Close();
                //udpclient = null;
                Console.WriteLine("udpclient.Close()");
            }
        }
        /// <summary>
        /// 打开接收数据线程
        /// </summary>
        /// <returns></returns>
        public void OpenReceive(Action<byte[], EndPoint> action)
        {
            ReceiveDataAction = action;
            if (!IsUdpcRecvStart)
            {
                IsUdpcRecvStart = true;
                if (udpclient != null)
                    udpclient.BeginReceive(ReceiveCallback, null);

                Console.WriteLine("udpclient.Open()");
            }
        }
        /// <summary>
        /// 接收数据回调
        /// </summary>
        /// <param name="ar"></param>
        private void ReceiveCallback(IAsyncResult ar)
        {
            IPEndPoint remoteIpep = null;
            byte[] buffer = null;
            try
            {
                buffer = udpclient.EndReceive(ar, ref remoteIpep);
                ReceiveDataAction?.Invoke(buffer, remoteIpep);
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
            }
            finally
            {
                if (IsUdpcRecvStart && udpclient != null)
                    udpclient.BeginReceive(ReceiveCallback, null);//递归
            }
        }
    }
}
