﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace RfidConsoleApp
{
    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct NRCLIENTSDK_CARD
    {
        [MarshalAs(UnmanagedType.U1)]
        public byte slot;           //分站号
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
        public byte[] cardNo;         //卡号
        [MarshalAs(UnmanagedType.U1)]
        public byte attr;
        public byte[] extData;
        [MarshalAs(UnmanagedType.U1)]
        public byte rssi;
    }

    [StructLayoutAttribute(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
    public struct Msg_Head
    {
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
        public byte[] ProtocolType;//"DSRF"

        public byte Version;

        public Msg_ID Msg_ID;

        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 2)]
        public byte[] CRC_Data;//16 位 CRC 校验码(消息长度和消息内容)
    }

    public struct Msg_Data
    {
        public Msg_Head Msg_Head;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 2)]
        public byte[] Data_Length;//消息内容长度
        public byte[] Data;
    }

    public enum Msg_ID : byte
    {
        Log_On_REQ = 1,//登录请求
        Log_On_RES = 2,//登录确认
        EleLabel_DATA = 3,//电子标签信息
        Heart_Beat_REQ = 4,//心跳请求
        Heart_Beat_RES = 5,//心跳确认
        Call_EleLabel_REQ = 6,//呼叫电子标签请求
        Call_EleLabel_RES = 7,//呼叫电子标签确认
        Call_EleLabel_Reject = 8,
        Call_EleLabel_MSG_REQ = 9,
        Call_EleLabel_MSG_RES = 10,
        Call_EleLabel_MSG_Reject = 11,
    }

    public enum LMsg_ID
    {
        Send_EleLabel_Data = 3,
        HandShake_REQ = 4,//Request
        HandShake_RES = 5,//Response
        Call_EleLabel_REQ = 6,//呼叫请求
        Call_EleLabel_RES = 7,
        Call_EleLabel_Reject = 8,
        Send_Message_REQ = 9,//发送短信请求
        Send_Message_RES = 10,
        Send_Message_Reject = 11,
    }

    public static class CRC
    {
        private static byte[] crc_16(byte[] data, int len)
        {
            ushort crc, i, j;
            if ((data == null) || (len == 0))
            {
                return new byte[] { 0, 0 };
            }
            crc = 0xffff;
            for (i = 0; i < len; i++)
            {
                crc = (ushort)(crc ^ (ushort)(data[i]));
                for (j = 0; j < 8; j++)
                {
                    if ((crc & 1) != 0)
                    {
                        crc = (ushort)((crc >> 1) ^ 40961);
                    }
                    else
                    {
                        crc = (ushort)(crc >> 1);
                    }
                }
            }
            byte hi = (byte)((crc & 0xFF00) >> 8);
            byte lo = (byte)(crc & 0x00FF);
            return new byte[] { hi, lo };
        }

        public static byte[] usMBCRC16(byte[] pucFrame, int usLen)
        {
            ushort ucCRCHi = 0xFF;
            ushort ucCRCLo = 0xFF;
            int iIndex;
            for (int i = 0; i < usLen; i++)
            {

                iIndex = ucCRCLo ^ pucFrame[i];
                ucCRCLo = (ushort)(ucCRCHi ^ aucCRCHi[iIndex]);
                ucCRCHi = aucCRCLo[iIndex];
            }
            ushort res = (ushort)(ucCRCHi << 8 | ucCRCLo);
            return BitConverter.GetBytes(res).Reverse().ToArray();
        }



        public static ushort[] aucCRCHi = new ushort[]
        {
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
            0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
    0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
    0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
    0x00, 0xC1, 0x81, 0x40
};

        public static ushort[] aucCRCLo = new ushort[]{
    0x00, 0xC0, 0xC1, 0x01, 0xC3, 0x03, 0x02, 0xC2, 0xC6, 0x06, 0x07, 0xC7,
    0x05, 0xC5, 0xC4, 0x04, 0xCC, 0x0C, 0x0D, 0xCD, 0x0F, 0xCF, 0xCE, 0x0E,
    0x0A, 0xCA, 0xCB, 0x0B, 0xC9, 0x09, 0x08, 0xC8, 0xD8, 0x18, 0x19, 0xD9,
    0x1B, 0xDB, 0xDA, 0x1A, 0x1E, 0xDE, 0xDF, 0x1F, 0xDD, 0x1D, 0x1C, 0xDC,
    0x14, 0xD4, 0xD5, 0x15, 0xD7, 0x17, 0x16, 0xD6, 0xD2, 0x12, 0x13, 0xD3,
    0x11, 0xD1, 0xD0, 0x10, 0xF0, 0x30, 0x31, 0xF1, 0x33, 0xF3, 0xF2, 0x32,
    0x36, 0xF6, 0xF7, 0x37, 0xF5, 0x35, 0x34, 0xF4, 0x3C, 0xFC, 0xFD, 0x3D,
    0xFF, 0x3F, 0x3E, 0xFE, 0xFA, 0x3A, 0x3B, 0xFB, 0x39, 0xF9, 0xF8, 0x38,
    0x28, 0xE8, 0xE9, 0x29, 0xEB, 0x2B, 0x2A, 0xEA, 0xEE, 0x2E, 0x2F, 0xEF,
    0x2D, 0xED, 0xEC, 0x2C, 0xE4, 0x24, 0x25, 0xE5, 0x27, 0xE7, 0xE6, 0x26,
    0x22, 0xE2, 0xE3, 0x23, 0xE1, 0x21, 0x20, 0xE0, 0xA0, 0x60, 0x61, 0xA1,
    0x63, 0xA3, 0xA2, 0x62, 0x66, 0xA6, 0xA7, 0x67, 0xA5, 0x65, 0x64, 0xA4,
    0x6C, 0xAC, 0xAD, 0x6D, 0xAF, 0x6F, 0x6E, 0xAE, 0xAA, 0x6A, 0x6B, 0xAB,
    0x69, 0xA9, 0xA8, 0x68, 0x78, 0xB8, 0xB9, 0x79, 0xBB, 0x7B, 0x7A, 0xBA,
    0xBE, 0x7E, 0x7F, 0xBF, 0x7D, 0xBD, 0xBC, 0x7C, 0xB4, 0x74, 0x75, 0xB5,
    0x77, 0xB7, 0xB6, 0x76, 0x72, 0xB2, 0xB3, 0x73, 0xB1, 0x71, 0x70, 0xB0,
    0x50, 0x90, 0x91, 0x51, 0x93, 0x53, 0x52, 0x92, 0x96, 0x56, 0x57, 0x97,
    0x55, 0x95, 0x94, 0x54, 0x9C, 0x5C, 0x5D, 0x9D, 0x5F, 0x9F, 0x9E, 0x5E,
    0x5A, 0x9A, 0x9B, 0x5B, 0x99, 0x59, 0x58, 0x98, 0x88, 0x48, 0x49, 0x89,
    0x4B, 0x8B, 0x8A, 0x4A, 0x4E, 0x8E, 0x8F, 0x4F, 0x8D, 0x4D, 0x4C, 0x8C,
    0x44, 0x84, 0x85, 0x45, 0x87, 0x47, 0x46, 0x86, 0x82, 0x42, 0x43, 0x83,
    0x41, 0x81, 0x80, 0x40
};


    }

    public static class PDU
    {
        static Msg_Data GetMsgHead_Buffer(Msg_ID id)
        {
            Msg_Data msg_data = new Msg_Data();
            msg_data.Msg_Head = new Msg_Head();
            var head = Encoding.ASCII.GetBytes("DSRF");
            msg_data.Msg_Head.ProtocolType = head;
            msg_data.Msg_Head.Version = Convert.ToByte(2);//协议版本2
            msg_data.Msg_Head.Msg_ID = id;
            return msg_data;
        }

        public static byte[] GetLogOn_Buffer()
        {
            var msg_data = GetMsgHead_Buffer(Msg_ID.Log_On_REQ);
            List<byte> datas = new List<byte>();
            var name = Encoding.ASCII.GetBytes("admin");
            byte[] n = new byte[16];
            name.CopyTo(n, 0);
            datas.AddRange(n);
            var pwd = Encoding.ASCII.GetBytes("888888");
            byte[] p = new byte[16];
            pwd.CopyTo(p, 0);
            datas.AddRange(p);
            ushort len = (ushort)datas.Count;
            msg_data.Data_Length = BitConverter.GetBytes(len).Reverse().ToArray();
            msg_data.Data = datas.ToArray();
            return CreateMsg_buffer(msg_data);
        }

        public static byte[] GetHeartBeat_Buffer()
        {
            var msg_data = GetMsgHead_Buffer(Msg_ID.Heart_Beat_REQ);
            msg_data.Data = null;
            msg_data.Data_Length = new byte[] { 0, 0 };
            return CreateMsg_buffer(msg_data);
        }


        public static byte[] GetCallTag_Buffer(int Slot, int id, int Mode, int seconds)
        {
            var c = BitConverter.GetBytes(id).Reverse().ToArray();
            var msg_data = GetMsgHead_Buffer(Msg_ID.Call_EleLabel_REQ);
            msg_data.Data_Length = new byte[] { 0, 7 };
            msg_data.Data = new byte[7];
            msg_data.Data[0] = (byte)Convert.ToInt32(Slot);
            Array.Copy(c, 0, msg_data.Data, 1, 4);
            msg_data.Data[5] = Convert.ToByte(seconds);
            msg_data.Data[6] = Convert.ToByte(Mode);//0:灯 1:灯+声音  2:red 4:green+red 255:停止

            return CreateMsg_buffer(msg_data);
        }

        public static byte[] GetSendShortMsg_Buffer(string smsg, int id)
        {
            if (string.IsNullOrEmpty(smsg))
            {
                return new byte[0];
            }
            var mdata = Encoding.Default.GetBytes(smsg);
            var label_id_data = BitConverter.GetBytes(Convert.ToInt32(id)).Reverse().ToArray();
            var msg_data = GetMsgHead_Buffer(Msg_ID.Call_EleLabel_MSG_REQ);
            //消息体
            msg_data.Data = new byte[124];
            msg_data.Data[0] = (byte)Convert.ToInt32(2);//分站号
            Array.Copy(label_id_data, 0, msg_data.Data, 1, 4);//标签号
            byte[] mbyte = new byte[] { 0, 0 };
            Array.Copy(mbyte, 0, msg_data.Data, 5, 2);
            msg_data.Data[7] = Convert.ToByte(DateTime.Now.Year - 2020);
            msg_data.Data[8] = Convert.ToByte(DateTime.Now.Month);
            msg_data.Data[9] = Convert.ToByte(DateTime.Now.Day);
            msg_data.Data[10] = Convert.ToByte(DateTime.Now.Hour);
            msg_data.Data[11] = Convert.ToByte(DateTime.Now.Minute);
            msg_data.Data[12] = Convert.ToByte(DateTime.Now.Second);

            if (mdata.Length > 110)
            {
                msg_data.Data[13] = Convert.ToByte(110);//短信内容长度
                Array.Copy(mdata, 0, msg_data.Data, 14, 110);
            }
            else
            {
                msg_data.Data[13] = Convert.ToByte(mdata.Length);//短信内容长度
                Array.Copy(mdata, 0, msg_data.Data, 14, mdata.Length);
            }
            msg_data.Data_Length = new byte[] { 0, Convert.ToByte(msg_data.Data.Length) };
            return CreateMsg_buffer(msg_data);
        }

        private static byte[] CreateMsg_buffer(Msg_Data msg_data)
        {
            List<byte> crcarray = new List<byte>();
            crcarray.AddRange(msg_data.Data_Length);
            if (msg_data.Data != null)
                crcarray.AddRange(msg_data.Data);
            var crcr = CRC.usMBCRC16(crcarray.ToArray(), crcarray.Count);
            msg_data.Msg_Head.CRC_Data = crcr;
            var sd = StructToBytes(msg_data.Msg_Head);
            var sdl = sd.ToList<byte>();
            sdl.AddRange(crcarray);
            return sdl.ToArray();
        }

        public static byte[] StructToBytes(object obj)
        {
            //得到结构体的大小
            int size = Marshal.SizeOf(obj);
            //创建byte数组
            byte[] bytes = new byte[size];
            //分配结构体大小的内存空间
            IntPtr structPtr = Marshal.AllocHGlobal(size);
            //将结构体拷到分配好的内存空间
            Marshal.StructureToPtr(obj, structPtr, false);
            //从内存空间拷到byte数组
            Marshal.Copy(structPtr, bytes, 0, size);
            //释放内存空间
            Marshal.FreeHGlobal(structPtr);
            //返回byte数组
            return bytes;
        }

        public static object BytesToStruct(byte[] bytes, Type type)
        {
            int size = Marshal.SizeOf(type);
            if (size > bytes.Length)
                return null;
            IntPtr structPtr = Marshal.AllocHGlobal(size); //分配结构大小的内存空间
            Marshal.Copy(bytes, 0, structPtr, size);       //将byte数组拷到分配好的内存空间
            object obj = Marshal.PtrToStructure(structPtr, type);
            Marshal.FreeHGlobal(structPtr);
            return obj;
        }
    }
}
